/**
 * @vue-async/resource-manager@1.0.3
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('vue'), require('@vue-async/utils')) :
  typeof define === 'function' && define.amd ? define(['exports', 'vue', '@vue-async/utils'], factory) :
  (global = global || self, factory(global.VueAsyncResourceManager = {}, global.Vue, global.VueAsyncUtils));
}(this, (function (exports, Vue, utils) { 'use strict';

  Vue = Vue && Object.prototype.hasOwnProperty.call(Vue, 'default') ? Vue['default'] : Vue;

  let currentInstance = null;
  /**
   * set current instance
   * @param instance Vue instance
   */
  function setCurrentInstance(instance) {
      currentInstance = instance;
  }
  const suspenseInstanceStack = [];
  let currentSuspenseInstance;
  /**
   * push suspense instance
   * @param instance Vue instance
   */
  function pushSuspenseInstance(instance) {
      currentSuspenseInstance = instance;
      suspenseInstanceStack.push(instance);
  }
  /**
   * pop suspense instance
   */
  function popSuspenseInstance() {
      suspenseInstanceStack.pop();
      return (currentSuspenseInstance = suspenseInstanceStack[suspenseInstanceStack.length - 1]);
  }

  const RESOLVED = 'resolved';
  const REJECTED = 'rejected';
  const COMPONENT_NAME = '__VueSuspense__';
  const del = (asyncFactory, error) => {
      const suspIns = asyncFactory.suspenseInstance;
      if (!suspIns) {
          return error("development" === 'production', 'No Suspense instance');
      }
      const asyncFactorys = suspIns.asyncFactorys;
      if (error) {
          suspIns.$emit(REJECTED, error);
          return;
      }
      asyncFactorys.delete(asyncFactory);
      if (asyncFactorys.size === 0) {
          suspIns.$emit(RESOLVED);
      }
  };
  const add = (asyncFactory) => {
      const suspIns = currentSuspenseInstance || asyncFactory.suspenseInstance;
      if (!suspIns) {
          return utils.error("development" === 'production', 'No Suspense instance');
      }
      const asyncFactorys = suspIns.asyncFactorys || (suspIns.asyncFactorys = new Set());
      if (suspIns.resolved) {
          suspIns.resolved = false;
          suspIns.setupLoading();
      }
      asyncFactorys.add(asyncFactory);
  };
  const has = (asyncFactory) => {
      const suspIns = currentSuspenseInstance || asyncFactory.suspenseInstance;
      if (!suspIns) {
          return utils.error("development" === 'production', 'No Suspense instance');
      }
      return suspIns.asyncFactorys && suspIns.asyncFactorys.has(asyncFactory);
  };
  var Suspense = Vue.extend({
      name: COMPONENT_NAME,
      props: {
          delay: {
              type: Number,
              default: 0,
          },
      },
      data() {
          return {
              resolved: false,
              rejected: false,
              displayLoading: false,
          };
      },
      methods: {
          setupLoading() {
              if (this.delay > 0) {
                  this._time = setTimeout(() => {
                      this.displayLoading = true;
                  }, this.delay);
              }
              else {
                  this.displayLoading = true;
              }
          },
          destoryLoading() {
              if (this._time) {
                  clearTimeout(this._time);
                  this._time = null;
              }
              this.displayLoading = false;
          },
      },
      created() {
          pushSuspenseInstance(this);
          this.promiser = new Promise((reslove, reject) => {
              this.$on(RESOLVED, () => {
                  this.destoryLoading();
                  this.resolved = true;
                  reslove();
              });
              this.$on(REJECTED, (err) => {
                  this.destoryLoading();
                  this.rejected = true;
                  reject(err);
              });
          });
          // start loading
          this.setupLoading();
      },
      mounted() {
          if (!this.asyncFactorys) {
              this.$emit(RESOLVED);
          }
      },
      updated() {
          if (!this.resolved)
              return;
          popSuspenseInstance();
      },
      render(h) {
          const isVisible = this.$options.suspense.mode === 'visible';
          const emptyVNode = this._e();
          const fallback = this.displayLoading ? this.$slots.fallback || [emptyVNode] : [emptyVNode];
          // The `children` is the real content to be rendered
          const children = this.$slots.default || [emptyVNode];
          let rendered;
          if (this.rejected && this.$slots.error) {
              rendered = createWrapper(h, this.$slots.error);
          }
          else {
              rendered = isVisible
                  ? createWrapper(h, this.resolved ? children : children.concat(fallback))
                  : createWrapper(h, [
                      // We need to render the children, but we should not show the rendered content.
                      createHiddenWrapper(h, children, this.resolved),
                      fallback,
                  ]);
          }
          return rendered;
      },
  });
  function createWrapper(h, children) {
      return h('div', {
          class: { 'vue-suspense-wrapper': true },
      }, children);
  }
  function createHiddenWrapper(h, tree, display) {
      return h('div', {
          style: { display: display ? 'block' : 'none' },
          class: { 'vue-suspense-hidden-wrapper': true },
      }, tree);
  }

  function findSuspenseInstance(ins) {
      let current = ins.$parent;
      while (current) {
          if (current.$options.name === COMPONENT_NAME) {
              return current;
          }
          else {
              current = current.$parent;
          }
      }
      return;
  }

  function Lazy(asyncFactory, props) {
      return Vue.extend({
          name: 'SuspenseLazy',
          props: props,
          created() {
              asyncFactory.suspenseInstance = currentSuspenseInstance || findSuspenseInstance(this);
              if (has(asyncFactory))
                  return;
              add(asyncFactory);
              if (asyncFactory.resolved) {
                  asyncFactory.$$waiter.then(() => {
                      del(asyncFactory);
                  });
                  return;
              }
              const promise = asyncFactory();
              asyncFactory.$$waiter = promise;
              promise
                  .then((C) => {
                  // Compatible ES Module
                  if (C.__esModule && C.default) {
                      C = C.default;
                  }
                  asyncFactory.resolved = C;
                  // Trigger update
                  this.$forceUpdate();
              })
                  .catch((err) => {
                  utils.error("development" === 'production', err);
                  del(asyncFactory, err);
              });
          },
          updated() {
              del(asyncFactory);
          },
          render(h) {
              // Fix context
              const slots = Object.keys(this.$slots)
                  .reduce((arr, key) => arr.concat(this.$slots[key] || []), [])
                  .map((vnode) => {
                  vnode.context = this._self;
                  return vnode;
              });
              return asyncFactory.resolved
                  ? h(asyncFactory.resolved, {
                      on: this.$listeners,
                      // If there is no props definition, fall back to `this.$attrs`
                      props: props ? this.$props : this.$attrs,
                      // Passthrough scopedSlots
                      scopedSlots: this.$scopedSlots,
                      attrs: this.$attrs,
                  }, slots)
                  : this._e();
          },
      });
  }

  Vue.mixin({
      created() {
          setCurrentInstance(this);
      },
  });
  //Vue.observable works in 2.6+
  function observable(data) {
      if (Vue.observable) {
          return Vue.observable(data);
      }
      return new Vue({
          data() {
              return data;
          },
      }).$data;
  }
  function CreateResource(fetchFactory, options) {
      const $res = observable({
          $$result: null,
          $$error: null,
          $$loading: false,
          $$loaded: false,
      });
      const currInstance = currentInstance;
      fetchFactory.suspenseInstance = currentSuspenseInstance
          ? currentSuspenseInstance
          : findSuspenseInstance(currInstance);
      const hasSuspenseInstance = !!fetchFactory.suspenseInstance;
      const resourceResult = {
          read(input) {
              // prevent
              if (options && options.prevent && $res.$$loading) {
                  return $res.$$promiser;
              }
              $res.$$loading = true;
              // Because we don't need caching, this is just a unique identifier,
              // and each call to .read() is a completely new request.
              const uniqueWrapFactory = (i) => fetchFactory(i);
              if (hasSuspenseInstance) {
                  // Establish a relationship between the fetchFactory and the current component instance
                  uniqueWrapFactory.suspenseInstance = fetchFactory.suspenseInstance;
                  add(uniqueWrapFactory);
              }
              // Start fetching asynchronous data
              const promise = ($res.$$promiser = uniqueWrapFactory(input));
              promise
                  .then((res) => {
                  // Trigger update
                  $res.$$result = options && options.onSuccess ? options.onSuccess(res) : res;
                  if (!$res.$$loaded) {
                      $res.$$loaded = true;
                  }
                  if (hasSuspenseInstance) {
                      del(uniqueWrapFactory);
                  }
              })
                  .catch((err) => {
                  utils.error("development" === 'production', err.message);
                  $res.$$error = options && options.onError ? options.onError(err) : err;
                  if (hasSuspenseInstance) {
                      del(uniqueWrapFactory, err);
                  }
              })
                  .finally(() => {
                  $res.$$loading = false;
              });
              return promise;
          },
          get $result() {
              return $res.$$result;
          },
          set $result(val) {
              $res.$$result = val;
          },
          get $error() {
              return $res.$$error;
          },
          set $error(val) {
              $res.$$error = val;
          },
          get $loading() {
              return $res.$$loading;
          },
          get $loaded() {
              return $res.$$loaded;
          },
          fork() {
              return CreateResource((i) => fetchFactory(i), options);
          },
      };
      return resourceResult;
  }

  function install(Vue, options = {}) {
      if (install.installed)
          return;
      install.installed = true;
      Vue.component('Suspense', Suspense);
      const opts = Object.assign({
          mode: 'visible',
      }, options);
      Vue.setSuspenseOptions = (options) => {
          Object.assign(opts, options);
      };
      Vue.mixin({
          created() {
              if (this.$options.name === COMPONENT_NAME) {
                  this.$options.suspense = opts;
              }
          },
      });
  }

  var index = {
      version: '1.0.3',
      install,
  };
  // Auto install if it is not done yet and `window` has `Vue`.
  // To allow users to avoid auto-installation in some cases,
  if (window && window.Vue) {
      window.Vue.use(install);
  }

  exports.Suspense = Suspense;
  exports.createReaourse = CreateResource;
  exports.default = index;
  exports.lazy = Lazy;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
